"""
===========
Chirper GUI
===========

This is the app associated with Chirper, which allows the user to
visualize and manipulate signals from different sources. Currently, it
is able to fetch data from the following sources:
 - Microphone
"""